;########################################################################
;#                                                                      #
;#                        ATN.LSP                                       #
;#                                                                      #
;#   CONTENTS:                                                          #
;#        Functions:  reject (atn), skip (atn) (always-true (atn char)  #
;#                    edge (start end test atn-action &key test-object  #
;#                          action-object)                              #
;#                    is-null (atn char), is-numeric (atn char)         #
;#                    not-numeric (atn char), is-alphabetic (atn char)  #
;#                    not-alphabetic (atn char), is-separator (atn char)#
;#                    is-punctuation (atn char), is-apostrophe(atn char)#
;#                    is-determiner (atn word), not-determiner(atn word)#
;#                    is-noun (atn word), is-verb (atn word),           #
;#                    is-pronoun (atn word)                             #
;#       Classes:   atn, edge                                           #
;#       Parameters:                                                    #
;#         atns:   number-atn, numbers-atn, word-atn, punct-atn,        #
;#                 recognize-words                                      #
;#                                                                      #
;#      Comment:  My lack of linguistic knwoledge prevents me from      #
;#                really understanding this file.  From what I can tell #
;#                the only thing it is used for is to find words in the #
;#                user's input                                          #
;#                                                                      #
;########################################################################

;;;--------------------------------------------------------
;;; Class definitions: atn and edge
;;;--------------------------------------------------------

(defclass atn ()
    ((edges 
       :initarg :edges
       :accessor edges
       :initform '())
     (current-state
       :initarg :current
       :accessor current
       :initform nil)
     (error-state
       :initarg :error-state
       :accessor error-state
       :initform 'error)
     (start-state
       :initarg :start-state
       :accessor start-state
       :initform 'start)
     (end-states
       :initarg :end-states
       :accessor end-states
       :initform '(end))
     (my-sequence
        :initarg :my-sequence
        :accessor my-sequence
        :initform '())
     (accumulator
       :initarg :accumulator
       :accessor accumulator
       :initform '())
     (index
       :initarg :index
       :accessor index
       :initform 0)))

(defclass edge ()
    ((start-node 
       :initarg :start
       :accessor start)
     (end-node
       :initarg :end
       :accessor end)
     (test
       :initarg :test
       :accessor test
       :initform '())
     (atn-action
       :initarg :atn-action
       :accessor atn-action
       :initform '())))

;;;--------------------------------------------------------
;;;  Atn methods
;;;--------------------------------------------------------

(defmethod reset ((self atn))
      (setf (current self) (start-state self))
      (setf (accumulator self) '())
      (setf (index self) 0))


(defmethod walk-string ((self atn) string)
    (when (typep string 'string)
       (setf string (string-to-list string)))
   (reset self)
   (sub-atn self string 0)
   (when  (success self)
      (accumulator self)))

(defmethod current-event ((self atn))  ;;; returns object currently pointed to by index
   (nth (index self) (my-sequence self)))

(defmethod input ((self atn))
   (block successful
   ;   (format t "~%   input index: ~a char:~a " (index self) (current-event self))
    (dolist (an-edge (edges self))
      (when (applicable an-edge self)
         
   ;      (format t "~%      method selected: ~s " (atn-action an-edge))
         (setf (current self) (end an-edge))
         (setf (index self) (funcall (atn-action an-edge) self))
         (return-from successful)))
    (setf (current self) (error-state self))))

(defmethod sub-atn ((self atn) string start-index)
   (reset self)
   (setf (my-sequence self) string)
   (setf (index self) start-index) 
  ; (format t "~%Subatn Index:~a~%" (index self))
   (loop until (or (success self) (failure self))
       do (input self)))

(defmethod success ((self atn))
   (member (current self) (end-states self)))   

(defmethod failure ((self atn))
   (equal (current self) (error-state self)))


;;;--------------------------------------------------------
;;; Edge methods
;;;--------------------------------------------------------          
(defmethod applicable ((self edge) atn)
   (and (eq (current atn) (start self))
        (funcall (test self) atn (current-event atn))))

;;;--------------------------------------------------------
;;;
;;; ATN Description Vocabulary
;;;
;;;   Note: Your ATN must have an edge for all possible
;;;         inputs. In addition to any inputs which you
;;;         define, the ATN must be able to handle the
;;;         nil input which represents the end of the
;;;         input sequence. (always-true or null will do this)
;;;
;;;   Predicates
;;;      
;;;      always-true
;;;      is-numeric
;;;      not-numeric
;;;      is-alphabetic
;;;      not-alphabetic
;;;      is-separator
;;;      is-null
;;;      succeeds  : needs lambda to take sub-atn arg
;;;
;;;   Actions
;;;      reject           - do not accept this char
;;;      skip             - skip over this char
;;;      collect-char          - consume char and collect to acc
;;;      collect-results  - run sub atn and collect results
;;;                         needs lambda for sub-atn arg
;;;
;;;--------------------------------------------------------     


;;;--------------------------------------------------------
;;; ATN Description Method Definitions
;;;--------------------------------------------------------          

(defun edge (start end test atn-action &key test-object action-object)
   (make-instance 'edge :start start :end end 
      :test test :atn-action atn-action ))


(defun is-null (atn char)
   (null char))

(defun is-numeric (atn char)
   (when char
   (and (char<= #\0 char)
        (char>= #\9 char))))

(defun not-numeric (atn char)
   (not (is-numeric atn char)))

(defun is-alphabetic (atn char)
   (when char
      (or
      (and (char<= #\a char)
           (char>= #\z char))   
      (and (char<= #\A char)
           (char>= #\Z char))
      (char= #\- char))))

(defun not-alphabetic (atn char)
   (not (is-alphabetic atn char)))

(defun is-separator (atn char)
   (not (or (is-numeric atn char)
            (is-alphabetic atn char))))

(defun is-punctuation (atn char)
   (when char
       (or (char= #\. char)
           (char= #\, char)
           (char= #\! char)
           (char= #\? char)
           (char= #\( char)
           (char= #\) char)
           (char= #\: char)
           (char= #\; char)
           (char= #\' char)
           (char= #\" char))))

(defun is-apostrophe (atn char)
   (char= #\' char))

(defmethod collect-char ((self atn))
    (setf (accumulator self)
          (append (accumulator self) (list (current-event self) )))
    (+ (index self) 1)) 

(defun succeeds (atn char sub-atn)
   (sub-atn sub-atn (my-sequence atn) (index atn))
 ;  (format t "   SUCCEEDS PRED:~a " (success sub-atn))
   (success sub-atn))
    

(defmethod collect-results ((self atn) sub-atn)
   (sub-atn sub-atn (my-sequence self) (index self))
   (setf (accumulator self) (append (accumulator self) (list (accumulator sub-atn))))
   (index sub-atn))

(defun reject (atn)  (index atn) )
(defun skip (atn) (+ (index atn) 1))

(defun always-true (atn char) t)

;;;--------------------------------------------------------
;;; Example atns
;;;--------------------------------------------------------

(defparameter number-atn 
(make-instance 'atn
   :edges (list
             (edge 'start 'error   #'is-null      #'reject)
             (edge 'start 'error #'not-numeric  #'reject)
             (edge 'start 'int   #'is-numeric   #'collect-char)
             (edge 'int   'int   #'is-numeric   #'collect-char)
             (edge 'int   'end   #'not-numeric  #'reject)
             ;(edge 'end   'end   #'always-true  #'reject) 
               )
   ))

(defparameter numbers-atn
      (make-instance 'atn
         :edges (list 
                   (edge 'start 'end #'is-null #'reject)
                   (edge 'start 'start 
                       #'(lambda (atn char) (succeeds atn char number-atn))
                       #'(lambda (atn) (collect-results atn number-atn)))
                   (edge 'end 'end #'always-true #'reject)
                   (edge 'start 'start #'not-numeric #'skip))))
                   

(defparameter word-atn
      (make-instance 'atn
         :edges (list
                   (edge 'start 'error #'is-null #'reject)
                   (edge 'start 'alpha #'is-apostrophe   #'collect-char)
                   (edge 'start 'error #'not-alphabetic  #'reject)
                   (edge 'start 'alpha #'is-alphabetic   #'collect-char)
                   (edge 'alpha 'alpha #'is-alphabetic   #'collect-char)
                   (edge 'alpha 'end   #'not-alphabetic  #'reject)
                  ; (edge 'end   'end   #'always-true    #'reject)  
                    )
   ))


(defparameter punct-atn
      (make-instance 'atn
         :edges (list 
                   (edge 'start 'end #'is-null #'reject)
                    (edge 'dots 'end #'is-null #'reject)
                   (edge 'start 'dots 
                      #'(lambda (atn char) (char= char #\.))
                      #'collect-char)
                   (edge 'dots 'dots
                      #'(lambda (atn char) (char= char #\.))
                      #'collect-char)
                   (edge 'dots 'end 
                      #'(lambda (atn char) (not (char= char #\.)))
                      #'reject)
                   (edge 'start 'end
                      #'is-punctuation
                      #'collect-char)
                  (edge 'end 'end #'always-true #'reject))))



(defparameter recognize-words
      (make-instance 'atn
         :edges (list
                   (edge 'start 'end #'is-null #'reject)
                   (edge 'start 'start 
                      #'(lambda (atn char) (succeeds atn char number-atn) )
                      #'(lambda (atn) (collect-results atn number-atn)))
                   (edge 'start 'start 
                      #'(lambda (atn char) (succeeds atn char word-atn) )
                      #'(lambda (atn) (collect-results atn word-atn)))
                   (edge 'start 'start
                      #'(lambda (atn char) (succeeds atn char punct-atn))
                      #'(lambda (atn) (collect-results atn punct-atn)))
                   (edge 'start 'start #'is-separator #'skip)
                   )))


(defun is-determiner (atn word)
   (and (not (unparsed-item word))
        (typep word 'determiner)))
(defun not-determiner (atn word)
   (not (is-determiner atn word)))
(defun is-noun (atn word)
    (and (not (unparsed-item word))
        (typep word 'noun)))
(defun is-verb (atn word )
   (and (not (unparsed-item word))
        (typep word 'verb)))
(defun is-pronoun (atn word)
   (and (not (unparsed-item word))
        (typepe word 'pronoun)))

;;; (defparameter noun-phrase-atn
;;;     ;;; This will only recognize noun phrases of the form det-noun or noun
;;;    (make-instance 'atn
;;;       :edges (list 
;;;                 (edge 'det 'end #'is-null #'reject)
;;;                 (edge 'start 'end #'is-null #'reject)
;;;                  (edge 'start 'end #'is-noun
;;;                    #'(lambda (atn)
;;;                         (setf (accumulator atn)
;;;                               (list (make-instance 'np :noun (current-event atn))))
;;;                         (+ (index atn) 1)))
;;;                 (edge 'start 'error #'not-determiner #'reject)
;;;                 (edge 'start 'det #'is-determiner
;;;                    #'(lambda (atn)
;;;                          (setf (accumulator atn) 
;;;                                (list (make-instance 'np :det (list (current-event 
;;;                                                                       atn)))))
;;;                         (+ (index atn) 1)))
;;;                 (edge 'det 'end #'is-noun
;;;                    #'(lambda (atn) 
;;;                            (setf (noun (first (accumulator atn))) (current-event atn))
;;;                         (+ (index atn) 1)))
;;;                  (edge 'det 'det #'is-determiner
;;;                    #'(lambda (atn) (setf (accumulator atn) 
;;;                                          (make-complex-det (last (accumulator atn))))
;;;                         (+ (index atn 1))))
;;;                 (edge 'det 'error #'always-true #'reject)
;;;                  )))
;;;                    
;;; (defparameter recognize-noun-phrases
;;;    (make-instance 'atn
;;;       :edges (list
;;;                 (edge 'start 'end #'is-null #'reject)
;;;                 (edge 'start 'start
;;;                    #'(lambda (atn word) (succeeds atn word noun-phrase-atn))
;;;                    #'(lambda (atn) 
;;;                         (collect-results atn noun-phrase-atn)
;;;                           ;  (+ (index atn) 1)
;;;                         ) )
;;;                 (edge 'start 'start #'always-true #'collect-char)
;;;               ;  (edge 'end 'end #'always-true #'reject)
;;;                 )))   
;;;                 
               
                
                







