;#######################################################################
;#                                                                     #
;#                      SCENARIO.LSP                                   #
;#                                                                     #
;#    CONTENTS:                                                        #
;#        Classes:  scenario                                           #
;#        Function:  choose-best-based-on-weights                      #
;#        Comment:  This file contains the scenario class and its      #
;#                  associated methods                                 #
;#                                                                     #
;#######################################################################


;*************************************************************
;*                                                           *
;*                  SCENARIO CLASS                           *
;*                                                           *
;*  name:  the name of the scenario                          *
;*  description:  the description                            *
;*  steps:  the list of steps in the scenario                *
;*  start-condition:  the conditions to tell if the scenario *
;*                    should be started                      *
;*  end-condition:  the conditions to tell if the scenario   *
;*                  is-finished                              *
;*  results: holds the results of testing if the scenario is *
;*           applicable or finished                          *
;*  parent:  the name of the scenario that it inherits steps *
;*           from                                            *
;*                                                           *
;*************************************************************

(defclass scenario ()
    ((name :type string 
        :initarg :name 
        :accessor name
        :initform "unnamed")
     (description :type string 
        :initarg :description 
        :accessor description
        :initform "a scenario")
     (steps :type list
        :initarg :steps
        :accessor steps
        :initform '() )
     (start-condition
      :initarg :my-condition
      :accessor my-condition
      :initform (spscorec :sc 2))
     (end-condition
      :initarg :end-condition
      :accessor end-condition
      :initform (lambdac #'(lambda (stuff)
                              (make-instance 'results))))
     (results
      :initarg :results
      :accessor results
      :initform '())
     (parent
      :initarg :parent
      :accessor parent
      :initform nil)))

(defmethod is-applicable ((self scenario) restaurant)
;;  Test the my-conditions of the scenario to see if it is applicable.
  (when debugp (format t "~%   inside scenario::is-applicable for ~a." self)
   (format t "~%   my-condition: ~a" (my-condition self))
)
 
   (setf (results self)
         (apply #'do-test (list (my-condition self)
                 restaurant self)))
   (when debugp (format t "~%   results after testing: ~a" (results self)))
   (success (results self)))

(defmethod is-finished ((self scenario) rest)
;; Test the end-conditions of the scenario to see if it is finished.
   (success (apply #'do-test (list (end-condition self) rest self))))

(defmethod do-next-steps ((self scenario) restaurant)
;; Evaluates the steps in the scenario to find out which to execute.
;; A list of applicable steps is first determined.  Then the best step is 
;; chosen and applied.  After the step has been applied the waiter's input
;; that made that step applicable is removed. 

   (let ((steps-to-evaluate '())
         (waiter  (waiter restaurant))
         (possible-steps '())
         (used-up-input '())
         (old-input (input (waiter restaurant)))
         (new-input '()))
      (when debugp (format t"~%   History of waiter: ~a" (history (waiter restaurant))))
     ; (format t "~%Time finding initial set of steps: ")
     ; (time 
         (dolist (a-step (steps self))                
         ;dolist binds a-step to each step in steps in turn.
         (when debugp (format t "~%  Considering ~a" a-step))    
         (when (is-applicable a-step restaurant)
           (when debugp (format t "~%  found an applicable step: ~a " a-step))
            (setf possible-steps (cons a-step possible-steps))))
      ;)
      
      (cond (possible-steps
               (when debugp 
                  (format t "~%  Looking for the best step to apply. Possible-steps:~% ~a"
                     possible-steps)
                  (format t "~% and their input-scores: ")
                  (loop for i in possible-steps do
                     (format t "~a  " (score (input-results i))))
                     
                  )
              ; (format t "~%Time spent choosing best steps: ")
              ; (time 
               (setf steps-to-evaluate (choose-best-steps self possible-steps restaurant))
              ; )
               (when debugp
                   (format t "~%   Found the best steps: ~a  starting execution" 
                               steps-to-evaluate)
                  )
               (loop for a-step in steps-to-evaluate do
                  (when (is-applicable a-step restaurant)
                     (do-step a-step restaurant)
                     (setf used-up-input 
                            (append (score (input-results a-step)) used-up-input))                           
                     (record waiter a-step)
                     (when debugp (format t "~%   Done ~a" a-step) 
                        (format t "~%   Waiter's history: ~a" (history waiter))
                        )
                        ))
               ;;;simplify waiter's input!!!
               ;;;add the score of the new step to the used-up-input and recalculate what 
               ;;;the new output would be. Set the waiter's output to this new output.
               (flatten used-up-input)
             ;;  (format t "~%used-up-input: ~a" used-up-input)
               (setf new-input '())
               (loop for index from (- (length old-input) 1) downto 0 do
                  (unless (member index used-up-input)
                     (push (nth index old-input) new-input)))
             ;;  (format t "~%new input: ~a" new-input)
               (setf (input (waiter restaurant)) new-input)
               steps-to-evaluate)
            (t nil))))

(defun choose-best-based-on-weights (choices)
  (setf best (car choices))
   (loop for a-step in choices do
     (when (> (total-score a-step) (total-score best))
       (setf best a-step)))
        best)


(defmethod choose-best-steps ((self scenario ) candidates restaurant)
;; If the waiter has input all the steps with input-condition are eliminated.
;; Then the step with the highest weight is chosen.

    (let ((candidate-list '())
          )
      
    ;  (format t "~%Choose-best-step called")
      (cond ((input (waiter restaurant)) 
             ;;;eliminate all candidates that do not use the input. 
             (loop for i in candidates do
                (unless (null (input-condition i))
                  (push i candidate-list)))
             
             (cond (candidate-list
                      (find-optimal-step-cover candidate-list))
                   (t  (list (choose-best-based-on-weights candidates)))
              
              ))
   
            (t 
               ;;;there is no input. Use the scores in the steps to compare them.
               (list (choose-best-based-on-weights candidates))
               ))
     )
                                                       
   )
      
(defun put-steps-in-order (candidate-list)
;; This is no longer used
   :documentation "This takes a list of steps and orders them according to the input they used. 
To calculate the order, take the average of the word-positions each step used. (eg. one that 
recognized words 1,3 and 5 would get a score of 9/3=3."
   (sort candidate-list 
      #'(lambda (x y) 
           (< (/ (reduce #'+ (score (input-results x))) (length (score (input-results x)))) 
              (/ (reduce #'+ (score (input-results y))) (length (score (input-results y))))))
      ))

(defmethod make-copy ((self scenario))
  :description "The copy is intended for use in the history list only. All components
   are shared except for the results, which are unique to the original and the copy."
  (make-instance 'scenario
         :name (name self)
	 :description (description self)
	 :steps (steps self)
	 :my-condition (my-condition self)
	 :results (make-copy (results self))))

(defmethod add-steps ((self scenario) new-step-list)
;;  Adds a list of steps to those inherited from the parents.
  (setf (steps self) (append new-step-list (steps (parent self)))))

(defmethod print-object ((self scenario) stream)
   (format stream "<~a>" (name self)))



 





